using VariationalAssimilation

using Plots
using StaticArrays
using DifferentialEquations
using DynamicalSystems
import PredefinedDynamicalSystems as pds

using Interpolations
using DelimitedFiles

#######

const hh_max_T = 300.0 
const noise_amplitude = 0.0 

#######


@inline nlss(V,θ,σ) = 0.5*(1+tanh(0.5*(V-θ)/σ))
@inline nltc(V,θ,σ,τ0,τ1) = τ0 + τ1*(1-tanh(0.5*(V-θ)/σ)^2)

function hhvf4d(u, p, drive)
    V,m,h,n = u

    Cm, Isa,
    EL, EK, ENa, 
    gL, gK, gNa,
    θm, σmss, σmtc, τm0, τm1,
    θh, σhss, σhtc, τh0, τh1,
    θn, σnss, σntc, τn0, τn1 = p

    Iapp = drive[1]

    dV = (1/Cm)*(gNa*m*m*m*h*(ENa - V) +
        gK*n*n*n*n*(EK - V) +
        gL*(EL - V) +
        Iapp/Isa)
    dm = (nlss(V,θm,σmss) - m)/nltc(V,θm,σmtc,τm0,τm1)
    dh = (nlss(V,θh,σhss) - h)/nltc(V,θh,σhtc,τh0,τh1)
    dn = (nlss(V,θn,σnss) - n)/nltc(V,θn,σntc,τn0,τn1)

    [dV,dm,dh,dn]
end

@sympy_register tanh
@sympy_register sqrt
@sympy_register exp

const D = 4;
const P = 23;

########################################################
# Generate synthetic data to perform twin experiment


function scale_timeseries(ts, minv, maxv)
    tsmax = maximum(ts)
    tsmin = minimum(ts)
    minv .+ (maxv - minv)*(ts .- tsmin)/(tsmax - tsmin) 
end

lor_sys = pds.lorenz(SA[-1.31;0.8;19.77])

const lor_max_T = 1600.0;
const lor_dt = 0.02;
lor_traj_rec, lor_trec = trajectory(lor_sys, lor_max_T; Δt = lor_dt)

lor_Iapp_interp(Iapp_tmax, Iapp_min, Iapp_max) = linear_interpolation(
    scale_timeseries(lor_trec, 0.0, Iapp_tmax),
    scale_timeseries(lor_traj_rec[:,1], Iapp_min, Iapp_max),
    extrapolation_bc = 0.0
    );


const gtp_4d = (Cm = 1.0,
    Isa = 1e-3,
    EL = -54.4,
    EK = -77.0,
    ENa = 50.0,
    gL = 0.3,
    gK = 20.0,
    gNa = 120.0,
    θm = -40.0,
    #σm = 2.5,
    σmss = 5.5,
    σmtc = 5.5,
    τm0 = 0.1,
    τm1 = 0.4,
    θh = -60.0,
    σhss = -7.5,
    σhtc = -7.5,
    τh0 = 1.0,
    τh1 = 7.0,
    θn = -55.0,
    σnss = 15.0,
    σntc = 15.0,
    τn0 = 1.0,
    τn1 = 5.0,
);

Iapp = lor_Iapp_interp(19000.0,-0.03, 0.025);

const hh_u0 = [-74.0, 0.0, 1.0, 0.0];
const hh_dt = 0.04;

hhvf4d_noise(A) = (u,p,t) -> [A, 0.0, 0.0, 0.0];

true_prob = ODEProblem(
    (u,p,t) -> hhvf4d(u,p,[Iapp(t)]), 
    hh_u0, 
    (0.0,hh_max_T), 
    collect(gtp_4d))

true_sol = solve(true_prob, saveat = hh_dt)

stoch_prob = SDEProblem(
    (u,p,t) -> hhvf4d(u,p,[Iapp(t)]), 
    hhvf4d_noise(noise_amplitude),
    hh_u0, 
    (0.0,hh_max_T), 
    collect(gtp_4d))

stoch_sol = solve(stoch_prob, saveat = hh_dt)

###############################3

knot_times = collect(0.0:hh_dt:hh_max_T);
midp_times = [(knot_times[i] + knot_times[i+1])/2.0 for i=1:(length(knot_times) - 1)];
all_times = sort(vcat(knot_times, midp_times))

Iapp_rec = Iapp.(all_times);
Vobs_rec = [stoch_sol(t)[1] for t in all_times];

u_bds = (V = (-130.0, 80.0, -65.0),
    m = (0.0, 1.0, 0.2),
    h = (0.0, 1.0, 0.8),
    n = (0.0, 1.0, 0.2),
)

p_bds = (Cm = (1.0, 1.0, 1.0),
    Isa = (1e-4, 1.0, 0.5),
    EL = (-75.0, -45.0, -65.0),
    EK = (-100.0, -75.0, -90.0),
    ENa = (40.0, 60.0, 50.0),
    gL = (0.0, 20.0, 1.0),
    gK = (0.0, 200.0, 5.0),
    gNa = (0.0, 600.0, 100.0),
    θm = (-100.0, -20.0, -40.0),
    σmss = (0.02, 100.0, 50.0),
    σmtc = (0.02, 100.0, 50.0),
    τm0 = (0.02, 10.0, 5.0),
    τm1 = (0.02, 10.0, 5.0),
    θh = (-100.0, -20.0, -40.0),
    σhss = (-100.0, -0.01, -50.0),
    σhtc = (-100.0, -0.01, -50.0),
    τh0 = (0.02, 50.0, 25.0),
    τh1 = (0.02, 50.0, 25.0),
    θn = (-100.0, -20.0, -40.0),
    σnss = (0.02, 100.0, 50.0),
    σntc = (0.02, 100.0, 50.0),
    τn0 = (0.02, 10.0, 5.0),
    τn1 = (0.02, 10.0, 5.0),
)

const num_drives = 1;
const observed_dims = [1];
const rec_dt = 0.02;

dpsem = DPSEProblemMeta(hhvf4d, D, P, 
    num_drives, Iapp_rec', 
    observed_dims, Vobs_rec', 
    rec_dt, 
    collect(u_bds), collect(p_bds))

vavf, vatrc, obs, bds = build_dpse_problem(dpsem, SimpsonHermite())

# vap is an NLPProblem which can be passed to both Ipopt and MadNLP
vap = VarAssimProblem(vavf, vatrc, obs, bds, init=:user)


##################################

# assim_p = stats.solution[end-P+1:end];
# assim_u0 = stats.solution[1:D]

# assim_prob = ODEProblem(
#     (u,p,t) -> hhvf4d(u,p,[Iapp(t)]), 
#     assim_u0, 
#     (0.0,450), 
#     assim_p)

# assim_sol = solve(assim_prob, saveat = hh_dt)