struct DPSEProblemMeta
    vf :: Function #signature vf(u,p,drive)
    D :: Int
    P :: Int

    num_drives :: Int
    drive_recordings :: Matrix #SC x num timepoints

    observed_dims :: Vector{Int}
    observation_recordings :: Matrix #length(observed_dims) x num timepoints

    rec_dt :: Float64 #sampling period of recordings

    state_bds :: Vector{Tuple}
    param_bds :: Vector{Tuple}
end

function build_dpse_problem(dpsem::DPSEProblemMeta, disc::SimpsonHermite)
    (;vf, D, P, 
        num_drives, drive_recordings, 
        observed_dims, observation_recordings,
        rec_dt,
        state_bds, param_bds) = dpsem

    R = num_intermediate(disc)
    @assert size(drive_recordings)[2] % (R+1) == 1 "Length of each drive recording must be 1 mod $(R+1)."
    @assert size(observation_recordings)[2] % (R+1) == 1 "Length of each observation recording must be 1 mod $(R+1)."
    @assert size(drive_recordings)[2] == size(observation_recordings)[2] "Lengths of drive and observation recordings do not match."
    N = (size(drive_recordings)[2] - 1) ÷ (R+1)

    NOBS = length(observed_dims)
    OC = 2*NOBS #odd indexes are control, even indexes are slopes
    SC = num_drives + NOBS

    function controlled_vf(u,oc,p,sc)
        out = dpsem.vf(u,p,sc[1:num_drives])
        for i in eachindex(observed_dims)
            out[observed_dims[i]] += oc[2*i-1]*(sc[num_drives + i] - u[observed_dims[i]])
        end
        out
    end

    vavf = VarAssimVectorField(controlled_vf, D, OC, P, SC)

    function residual_with_ctrl(vavf::VarAssimVectorField, ublock, ocblock, p, scblock, dt, λblock)
        disc_cons = timestep_residual(disc, vavf, ublock, ocblock, p, scblock, dt)
        ctrl_cons = [ocblock[2][2*i-1] - ((ocblock[1][2*i-1] + ocblock[3][2*i-1])/2.0 + (dt/8.0)*(ocblock[1][2*i] - ocblock[3][2*i])) for i=1:NOBS]
        #This format of ctrl constraints only works for SimpsonHermite

        λblock .* vcat(disc_cons, ctrl_cons)
    end

    NC = num_constraints(disc, D) + NOBS

    function dpse_timestep_cost(vavf::VarAssimVectorField, ublock, ocblock, p, scblock, dt, λblock)
        cost = 0.0

        for j=1:(R+2)
            if j == 1 || j == R+2
                mult = 0.5
            else
                mult = 1.0
            end
            for i=1:NOBS
                cost_term = (ublock[j][observed_dims[i]] - scblock[j][num_drives+i])^2 + ocblock[j][2*i-1]^2 + ocblock[j][2*i]^2
                cost += mult*cost_term
            end
        end

        cost*dt/2.0
    end

    vatrc = VarAssimTimestepResidualCost(residual_with_ctrl, dpse_timestep_cost, R, NC)

    oc_bds = repeat([(0.0, 1.0, 0.5), (-1.0, 1.0, 0.0)], NOBS) #bounds for ctrl var and ctrl slopes
    bds = VarAssimBounds(state_bds, oc_bds, param_bds)

    knot_times = range(start = 0.0, step = (R+1)*rec_dt, length = N+1)
    SC_rec = vec(vcat(drive_recordings, observation_recordings))
    @assert length(SC_rec) == SC*(N*(R+1)+1)
    obs = VarAssimObservations(knot_times, SC_rec)

    vavf, vatrc, obs, bds
end

