import JuliaFormatter

function sp_cse_timestep_to_julia(vavf, sp_cse_intermediate, sp_cse_out)
    intermediate_length = pyconvert(Int, sp_cse_intermediate.__len__())
    out_length = pyconvert(Int, sp_cse_out.__len__())
    
    intermediate_eq_str = ""
    out_eq_str = ""
    
    for i=0:(intermediate_length - 1)
        lhs = pyconvert(String, sp.printing.julia.julia_code(sp_cse_intermediate[i][0]))
        rhs = pyconvert(String, sp.printing.julia.julia_code(sp_cse_intermediate[i][1]))
        
        intermediate_eq_str *= "$(lhs) = $(rhs)\n"
    end
    
    for i=0:(out_length - 1)
        out_val = pyconvert(String,sp.printing.julia.julia_code(sp_cse_out[i]))
        out_eq_str *= "OUTPUT[$(i+1)] += $(out_val)\n"
    end
    
    find_pats = [Regex(raw"u_(\d+)_(\d+)"), 
        Regex(raw"oc_(\d+)_(\d+)"), 
        Regex(raw"p_(\d+)"), 
        Regex(raw"sc_(\d+)_(\d+)"), 
        Regex(raw"lambda_(\d+)")]
    
    (;D, OC, SC) = vavf
    
    repl_pats = [SubstitutionString("uocblock[(\\1 - 1)*$(D+OC) + \\2]"), 
        SubstitutionString("uocblock[(\\1 - 1)*$(D+OC) + $(D) + \\2]"), 
        SubstitutionString("p[\\1]"), 
        SubstitutionString("scblock[(\\1 - 1)*$(SC) + \\2]"), 
        SubstitutionString("λblock[\\1]")]
    
    intermediate_eq_str = replace(intermediate_eq_str, [find_pats[i] => repl_pats[i] for i in eachindex(find_pats)]...)
    out_eq_str = replace(out_eq_str, [find_pats[i] => repl_pats[i] for i in eachindex(find_pats)]...)
    
    funcstr = intermediate_eq_str*out_eq_str
    
    funcstr
end

function cse_timestep_str_to_inner_loop_str(timestep_params, funcstr)
    (;D, OC, P, SC, R, NC) = timestep_params
    
    find_pats = [Regex(raw"uocblock\[([0-9+\-*/()\s]*)\]"),
        Regex(raw"p\[([0-9+\-*/()\s]*)\]"),
        Regex(raw"scblock\[([0-9+\-*/()\s]*)\]"),
        Regex(raw"λblock\[([0-9+\-*/()\s]*)\]"),
        Regex(raw"dt")]
    
    repl_pats = [SubstitutionString("opt_x[($(D)+$(OC))*($(R)+1)*(t-1) + \\1]"),
        SubstitutionString("opt_x[($(D)+$(OC))*(($(R)+1)*N+1) + \\1]"),
        SubstitutionString("SC_rec[$(SC)*($(R)+1)*(t-1) + \\1]"),
        SubstitutionString("λ[$(NC)*(t-1) + \\1]"),
        SubstitutionString("(knot_times[t+1] - knot_times[t])")]
    
    replace(funcstr, [find_pats[i] => repl_pats[i] for i in eachindex(find_pats)]...)
end

function sympy_workload(vavf, vatrc)
    (;D, OC, P, SC) = vavf
    (;R, NC) = vatrc
    
    @info "Begin SymPy workload. Making symbolic variables with SymPy..."
    
    ublock = [Py[sp.symbols("u_$(t)_$(j)") for j=1:D] for t=1:(R+2)]
    ocblock = [Py[sp.symbols("oc_$(t)_$(j)") for j=1:OC] for t=1:(R+2)]
    uocblock = [vcat(ublock[t], ocblock[t]) for t=1:(R+2)]

    pblock = Py[sp.symbols("p_$(j)") for j=1:P]
    scblock = [Py[sp.symbols("sc_$(t)_$(j)") for j=1:vavf.SC] for t=1:(R+2)]
    lambdablock = Py[sp.symbols("lambda_$(j)") for j=1:NC];

    dt = sp.symbols("dt")
    obj_factor = sp.symbols("obj_factor");
    
    @info "Building timestep cost, residual, and Lagrangian expressions..."
    
    sp_timestep_cost = vatrc.timestep_cost(vavf, ublock, ocblock, pblock, scblock, dt, lambdablock);
    sp_timestep_residual = vatrc.timestep_residual(vavf, ublock, ocblock, pblock, scblock, dt, lambdablock);
    sp_timestep_lagrangian = obj_factor*vatrc.timestep_cost(vavf, ublock, ocblock, pblock, scblock, dt, lambdablock) + sum(sp_timestep_residual);

    sp_deriv_vars = vcat(vcat(uocblock...), pblock); #variables to differentiate wrt -- uocblock and p only
    
    @info "Differentiating timestep cost, residual, and Lagrangian expressions..."
    
    sp_cost_grad = sp.Matrix([sp_timestep_cost]).jacobian(sp_deriv_vars);
    sp_res_jac = sp.Matrix(sp_timestep_residual).jacobian(sp_deriv_vars);
    sp_lagr_hess = sp.hessian(sp_timestep_lagrangian, sp_deriv_vars);
    
    @info "Computing timestep sparsity structures for cost gradient, residual Jacobian, and Hessian of Lagrangian..."

    sp_cost_grad_struct = Int64[]
    sp_cost_grad_val = Py[]

    sp_res_jac_struct = Tuple{Int64,Int64}[]
    sp_res_jac_val = Py[]

    sp_lagr_hess_struct = Tuple{Int64,Int64}[]
    sp_lagr_hess_val = Py[]

    for j in eachindex(sp_deriv_vars)
        cost_grad_entry = sp_cost_grad[j-1]
        if pyconvert(Bool, cost_grad_entry == 0)
            nothing
        else
            push!(sp_cost_grad_struct, j)
            push!(sp_cost_grad_val, cost_grad_entry)
        end
    end

    for j in eachindex(sp_deriv_vars)
        for c = 1:vatrc.NC
            jac_entry = sp_res_jac[c-1, j-1]
            if pyconvert(Bool, jac_entry == 0)
                nothing
            else
                push!(sp_res_jac_struct, (c,j))
                push!(sp_res_jac_val, jac_entry)
            end
        end
    end

    for j in eachindex(sp_deriv_vars)
        for i in j:length(sp_deriv_vars)
            hess_entry = sp_lagr_hess[i-1,j-1]
            if pyconvert(Bool, hess_entry == 0)
                nothing
            else
                push!(sp_lagr_hess_struct, (i,j))
                push!(sp_lagr_hess_val, hess_entry)
            end
        end
    end

    NNZ_G = length(sp_cost_grad_struct);
    NNZ_J = length(sp_res_jac_struct);
    NNZ_H = length(sp_lagr_hess_struct);
    
    #Each nonzero Hessian entry in the full problem is one of two types: either it has contribution from only one timestep or it has contribution from two consecutive timesteps
    #In the latter case, it must be duoc_duoc or dp_duoc for uoc at a knot time
    #So to facilitate filling the problem Hessian efficiently, we find these entries beforehand and plan to sum the twofers via a "stride overlap"

    @info "Reordering nonzero Hessian entries to facilitate assembly of problem Hessian..."

    hess_entry_dict = Dict(sp_lagr_hess_struct[i] => i for i=1:NNZ_H);

    dparam_dparam_idxs = findall(x -> x[1] > (D+OC)*(R+2) && x[2] > (D+OC)*(R+2), sp_lagr_hess_struct)
    duocp_dintermediate_idxs = findall(x -> (D+OC+1) <= x[2] <= (D+OC)*(R+1), sp_lagr_hess_struct)

    dleftknot_dleftknot_idxs = findall(x -> x[1]∈1:(D+OC) && x[2]∈1:(D+OC), sp_lagr_hess_struct)
    dparam_dleftknot_idxs = findall(x -> x[1]∈((D+OC)*(R+2)+1):((D+OC)*(R+2)+P) && x[2]∈1:(D+OC), sp_lagr_hess_struct)

    drightknot_drightknot_idxs = findall(x -> x[1]∈((D+OC)*(R+1)+1):((D+OC)*(R+2)) && x[2]∈((D+OC)*(R+1)+1):((D+OC)*(R+2)), sp_lagr_hess_struct)
    dparam_drightknot_idxs = findall(x -> x[1]∈((D+OC)*(R+2)+1):((D+OC)*(R+2)+P) && x[2]∈((D+OC)*(R+1)+1):((D+OC)*(R+2)), sp_lagr_hess_struct);

    @assert length(dparam_dparam_idxs)+length(duocp_dintermediate_idxs)+length(dleftknot_dleftknot_idxs) + length(dparam_dleftknot_idxs) + length(drightknot_drightknot_idxs) + length(dparam_drightknot_idxs) == length(sp_lagr_hess_struct)

    dleftknot_dleftknot_pairs = sp_lagr_hess_struct[dleftknot_dleftknot_idxs]
    dparam_dleftknot_pairs = sp_lagr_hess_struct[dparam_dleftknot_idxs]

    drightknot_drightknot_pairs = sp_lagr_hess_struct[drightknot_drightknot_idxs]
    dparam_drightknot_pairs = sp_lagr_hess_struct[dparam_drightknot_idxs];

    drightknot_drightknot_mated = [x for x in drightknot_drightknot_pairs if (x[1] - (R+1)*(D+OC), x[2] - (R+1)*(D+OC))∈dleftknot_dleftknot_pairs]
    dleftknot_dleftknot_mated = [(x[1] - (R+1)*(D+OC), x[2] - (R+1)*(D+OC)) for x in drightknot_drightknot_mated]

    dparam_drightknot_mated = [x for x in dparam_drightknot_pairs if (x[1], x[2] - (R+1)*(D+OC))∈dparam_dleftknot_pairs]
    dparam_dleftknot_mated = [(x[1], x[2] - (R+1)*(D+OC)) for x in dparam_drightknot_mated]

    drightknot_drightknot_mated_idxs = [hess_entry_dict[x] for x in drightknot_drightknot_mated]
    dleftknot_dleftknot_mated_idxs= [hess_entry_dict[x] for x in dleftknot_dleftknot_mated]

    dparam_drightknot_mated_idxs = [hess_entry_dict[x] for x in dparam_drightknot_mated]
    dparam_dleftknot_mated_idxs = [hess_entry_dict[x] for x in dparam_dleftknot_mated]

    mated_right_idxs = vcat(drightknot_drightknot_mated_idxs, dparam_drightknot_mated_idxs)
    mated_left_idxs = vcat(dleftknot_dleftknot_mated_idxs, dparam_dleftknot_mated_idxs);

    remaining_idxs = [x for x in 1:NNZ_H if !(x in vcat(mated_left_idxs, mated_right_idxs, dparam_dparam_idxs))]

    final_perm = vcat(mated_left_idxs, remaining_idxs, mated_right_idxs, dparam_dparam_idxs)

    @assert sort(final_perm) == 1:NNZ_H

    fin_sp_lagr_hess_struct = sp_lagr_hess_struct[final_perm];
    fin_sp_lagr_hess_val = sp_lagr_hess_val[final_perm];

    NZ_H_PP = length(dparam_dparam_idxs)
    NZ_H_M = length(mated_right_idxs)
    NZ_H_REM = length(remaining_idxs)

    @assert NZ_H_M + NZ_H_REM + NZ_H_M + NZ_H_PP == NNZ_H
    @assert Set(fin_sp_lagr_hess_struct[end-NZ_H_PP+1:end]) == Set(sp_lagr_hess_struct[dparam_dparam_idxs])
    @assert Set(fin_sp_lagr_hess_struct[1:NZ_H_M]) == Set(sp_lagr_hess_struct[mated_left_idxs])
    @assert Set(fin_sp_lagr_hess_struct[NZ_H_M+NZ_H_REM+1:2*NZ_H_M+NZ_H_REM]) == Set(sp_lagr_hess_struct[mated_right_idxs])
    
    @info "Performing common subexpression elimination (CSE) with SymPy..."

    sp_timestep_cost_cse_intermediate, sp_timestep_cost_cse_out = sp.cse(sp_timestep_cost)
    sp_cost_grad_cse_intermediate, sp_cost_grad_cse_out = sp.cse(sp_cost_grad_val);

    sp_timestep_residual_cse_intermediate, sp_timestep_residual_cse_out = sp.cse(sp_timestep_residual)
    sp_res_jac_cse_intermediate, sp_res_jac_cse_out = sp.cse(sp_res_jac_val);
    sp_lagr_hess_cse_intermediate, sp_lagr_hess_cse_out = sp.cse(fin_sp_lagr_hess_val);
    
    @info "Converting results of CSE to Julia strings..."

    timestep_cost_str = sp_cse_timestep_to_julia(vavf, sp_timestep_cost_cse_intermediate, sp_timestep_cost_cse_out)
    cost_grad_str = sp_cse_timestep_to_julia(vavf, sp_cost_grad_cse_intermediate, sp_cost_grad_cse_out)

    timestep_res_str = sp_cse_timestep_to_julia(vavf, sp_timestep_residual_cse_intermediate, sp_timestep_residual_cse_out)
    res_jac_str = sp_cse_timestep_to_julia(vavf, sp_res_jac_cse_intermediate, sp_res_jac_cse_out)
    lagr_hess_str = sp_cse_timestep_to_julia(vavf, sp_lagr_hess_cse_intermediate, sp_lagr_hess_cse_out);

    timestep_cse_dict = Dict(:cost => timestep_cost_str, 
        :grad => cost_grad_str,
        :res => timestep_res_str,
        :jac => res_jac_str,
        :hess => lagr_hess_str)
    
    timestep_struct_dict = Dict(:grad => sp_cost_grad_struct,
        :jac => sp_res_jac_struct,
        :hess => fin_sp_lagr_hess_struct)
    
    timestep_params = (D = D, 
        OC = OC, 
        P = P, 
        SC = SC, 
        NC = NC, 
        R = R, 
        NNZ_G = NNZ_G,
        NNZ_J = NNZ_J,
        NNZ_H = NNZ_H,
        NZ_H_PP = NZ_H_PP,
        NZ_H_M = NZ_H_M,
        NZ_H_REM = NZ_H_REM)
        
    timestep_cse_dict, timestep_struct_dict, timestep_params
end

function timestep_sparsity_to_problem_sparsity(timestep_struct_dict, timestep_params)
    (;D, OC, P, SC, NC, R, NNZ_G, NNZ_J, NNZ_H, NZ_H_PP, NZ_H_M, NZ_H_REM) = timestep_params

    @info "Constructing problem sparsity callbacks..."
    
    res_jac_rows = String[]
    res_jac_cols = String[]
    
    for k=1:NNZ_J
        i,j = timestep_struct_dict[:jac][k]
        push!(res_jac_rows, "$(i) + (t-1)*$(NC)")
        if j <= (D+OC)*(R+2)
            prob_j = "$(D+OC)*$(R+1)*(t-1) + $(j)"
        else
            prob_j = "$(D+OC)*($(R+1)*N+1) + $(j) - $(D+OC)*$(R+2)"
        end
        push!(res_jac_cols, prob_j)
    end
    
    problem_jac_struct_str = "function jac_struct_fill!(rows, cols, N)\n\tfor t=1:N\n\t@inbounds begin"

    for k=1:NNZ_J
        rowstr = "rows[$(NNZ_J)*(t-1)+$(k)] = "*res_jac_rows[k]
        colstr = "cols[$(NNZ_J)*(t-1)+$(k)] = "*res_jac_cols[k]
        problem_jac_struct_str = problem_jac_struct_str*"\n\t\t"*rowstr*"\n\t\t"*colstr
    end

    problem_jac_struct_str *= "\n\t\tend\n\tend\nreturn nothing;\nend"
    
    lagr_hess_rows = String[]
    lagr_hess_cols = String[]

    for k=1:NNZ_H
        i,j = timestep_struct_dict[:hess][k]
        if i <= (D+OC)*(R+2) && j <= (D+OC)*(R+2)
            prob_i = "$(D+OC)*$(R+1)*(t-1) + $(i)"
            prob_j = "$(D+OC)*$(R+1)*(t-1) + $(j)" 
        elseif i > (D+OC)*(R+2) && j <= (D+OC)*(R+2)
            prob_i = "$(D+OC)*($(R+1)*N+1) + $(i) - $(D+OC)*$(R+2)"
            prob_j = "$(D+OC)*$(R+1)*(t-1) + $(j)"
        else
            prob_i = "$(D+OC)*($(R+1)*N+1) + $(i) - $(D+OC)*$(R+2)"
            prob_j = "$(D+OC)*($(R+1)*N+1) + $(j) - $(D+OC)*$(R+2)"
        end
        push!(lagr_hess_rows, prob_i)
        push!(lagr_hess_cols, prob_j)
    end

    problem_hess_struct_str = "function hess_struct_fill!(rows, cols, N)\n\tfor t=1:1\n\t@inbounds begin"

    #new uoc_uoc and uoc_p entries from first timestep
    for k=1:(NZ_H_M + NZ_H_REM + NZ_H_M)
        rowstr = "rows[$(k)] = "*lagr_hess_rows[k]
        colstr = "cols[$(k)] = "*lagr_hess_cols[k]
        problem_hess_struct_str *= "\n\t\t"*rowstr*"\n\t\t"*colstr
    end

    problem_hess_struct_str *= "\n\t\tend\n\tend\n"
    problem_hess_struct_str *= "\nfor t=2:N\n\t@inbounds begin"

    #new uoc_uoc and uoc_p entries from timesteps 2 to N
    for k=1:(NZ_H_REM + NZ_H_M)
        rowstr = "rows[$(2*NZ_H_M+NZ_H_REM) + (t-2)*$(NZ_H_REM + NZ_H_M) + $(k)] ="*lagr_hess_rows[NZ_H_M+k]
        colstr = "cols[$(2*NZ_H_M+NZ_H_REM) + (t-2)*$(NZ_H_REM + NZ_H_M) + $(k)] ="*lagr_hess_cols[NZ_H_M+k]
        problem_hess_struct_str *= "\n\t\t"*rowstr*"\n\t\t"*colstr
    end

    problem_hess_struct_str *= "\n\t\tend\n\tend\n\n@inbounds begin"

    #p_p entries (contribution from all timesteps)
    for k=1:NZ_H_PP
        rowstr = "rows[N*$(NZ_H_M+NZ_H_REM) + $(NZ_H_M) + $(k)] ="*lagr_hess_rows[2*NZ_H_M+NZ_H_REM+k]
        colstr = "cols[N*$(NZ_H_M+NZ_H_REM) + $(NZ_H_M) + $(k)] ="*lagr_hess_cols[2*NZ_H_M+NZ_H_REM+k]
        problem_hess_struct_str *= "\n\t"*rowstr*"\n\t"*colstr
    end

    problem_hess_struct_str *= "\n\tend\nreturn nothing;\nend";
    
    #jac_struct_callback = Meta.parse(problem_jac_struct_str)
    #hess_struct_callback = Meta.parse(problem_hess_struct_str)
    
    #jac_struct_callback, hess_struct_callback
    problem_jac_struct_str, problem_hess_struct_str
end

function timestep_cse_to_problem_derivatives(timestep_cse_dict, timestep_struct_dict, timestep_params)
    (;D, OC, P, SC, NC, R, NNZ_G, NNZ_J, NNZ_H, NZ_H_PP, NZ_H_M, NZ_H_REM) = timestep_params

    inner_loop_dict = Dict(k => (cse_timestep_str_to_inner_loop_str(timestep_params, timestep_cse_dict[k])) for k in keys(timestep_cse_dict))

    @info "Fine-tuning inner loops for derivative callbacks..."
    
    #Timestep cost
    inner_loop_dict[:cost] = replace(inner_loop_dict[:cost], Regex(raw"OUTPUT\[([0-9+\-*/()\s]*)\]") => SubstitutionString("Obj[\\1]"))
    
    #Timestep cost gradient
    for j = 1:NNZ_G
        grad_entry = timestep_struct_dict[:grad][j]
        if grad_entry > (D+OC)*(R+2)
            corr_param = grad_entry - (D+OC)*(R+2)
            replacement = "Grad[$(D+OC)*($(R+1)*N+1) + $(corr_param)]"
        else
            replacement = "Grad[$(D+OC)*$(R+1)*(t-1) + $(grad_entry)]"
        end
        inner_loop_dict[:grad] = replace(inner_loop_dict[:grad], "OUTPUT[$(j)]" => replacement)
    end
    
    #Timestep residual
    inner_loop_dict[:res] = replace(inner_loop_dict[:res], Regex(raw"OUTPUT\[([0-9+\-*/()\s]*)\]") => SubstitutionString("Cons[$(NC)*(t-1) + \\1]"))

    #Timestep residual jacobian
    inner_loop_dict[:jac] = replace(inner_loop_dict[:jac], Regex(raw"OUTPUT\[([0-9+\-*/()\s]*)\]") => SubstitutionString("Jac[$(NNZ_J)*(t-1) + \\1]"))

    #Timestep Hessian
    for j = 1:NNZ_H
        if j > 2*NZ_H_M+NZ_H_REM
            corr_dpp = j - 2*NZ_H_M - NZ_H_REM
            replacement = "Hess[N*$(NZ_H_M + NZ_H_REM) + $(NZ_H_M) + $(corr_dpp)]"
        else
            replacement = "Hess[$(NZ_H_M + NZ_H_REM)*(t-1) + $(j)]"
        end
        inner_loop_dict[:hess] = replace(inner_loop_dict[:hess], "OUTPUT[$(j)]" => replacement)
    end
    
    @info "Constructing final derivative callbacks..."
    
    cost_str = """function cost_fill!(Obj, opt_x, λ, SC_rec, knot_times, N)
        fill!(Obj, 0.0)

        @inbounds for t = 1:N
            $(inner_loop_dict[:cost])
        end

    return nothing;
    end"""

    grad_str = """function grad_fill!(Grad, opt_x, λ, SC_rec, knot_times, N)
        fill!(Grad, 0.0)

        @inbounds for t = 1:N
            $(inner_loop_dict[:grad])
        end
    end"""

    res_str = """function cons_fill!(Cons, opt_x, λ, SC_rec, knot_times, N)
        fill!(Cons, 0.0)

        @inbounds for t = 1:N
            $(inner_loop_dict[:res])
        end

    return nothing;
    end"""

    jac_str = """function jac_fill!(Jac, opt_x, λ, SC_rec, knot_times, N)
        fill!(Jac, 0.0)

        @inbounds for t = 1:N
            $(inner_loop_dict[:jac])
        end

    return nothing;
    end"""

    hess_str = """function hess_fill!(Hess, opt_x, λ, obj_factor, SC_rec, knot_times, N)
        fill!(Hess, 0.0)

        @inbounds for t = 1:N
            $(inner_loop_dict[:hess])
        end

    return nothing;
    end""";
    
    # cost_callback = Meta.parse(cost_str)
    # grad_callback = Meta.parse(grad_str)
    # res_callback = Meta.parse(res_str)
    # jac_callback = Meta.parse(jac_str)
    # hess_callback = Meta.parse(hess_str)
    
    # cost_callback, grad_callback, res_callback, jac_callback, hess_callback
    cost_str, grad_str, res_str, jac_str, hess_str
end

function callback_module_str(vavf, vatrc)
    timestep_cse_dict, timestep_struct_dict, timestep_params = sympy_workload(vavf, vatrc)

    cost_str, grad_str, res_str, jac_str, hess_str = timestep_cse_to_problem_derivatives(timestep_cse_dict, timestep_struct_dict, timestep_params)
    jac_struct_str, hess_struct_str = timestep_sparsity_to_problem_sparsity(timestep_struct_dict, timestep_params)
    
    module_str = "module NLPCallbacks\n\n const timestep_params = $(timestep_params) \n\n"

    for x_str in [res_str, cost_str, grad_str, jac_str, jac_struct_str, hess_str, hess_struct_str]
        module_str *= "$(x_str)\n\n"
    end

    module_str *= "end"

    JuliaFormatter.format_text(module_str)
end


# function generate_ipopt_callbacks(vavf, vatrc; str = false)
#     timestep_cse_dict, timestep_struct_dict, timestep_params = sympy_workload(vavf, vatrc)
    
#     cost_str, grad_str, res_str, jac_str, hess_str = timestep_cse_to_problem_derivatives(timestep_cse_dict, timestep_struct_dict, timestep_params)
#     jac_struct_str, hess_struct_str = timestep_sparsity_to_problem_sparsity(timestep_struct_dict, timestep_params)
    
#     if str
#         return cost_str, grad_str, res_str, jac_str, hess_str, jac_struct_str, hess_struct_str, timestep_params
#     else
#         Meta.parse.([cost_str, grad_str, res_str, jac_str, hess_str, jac_struct_str, hess_struct_str])..., timestep_params
#     end
# end

