using NLPModels

mutable struct VarAssimProblem <: AbstractNLPModel{Float64,Vector{Float64}}
    obs :: VarAssimObservations
    bds :: VarAssimBounds
    
    callback_module :: Module
            
    Obj_cache :: Vector{Float64} #obj in callback_module is in-place, but NLPModels wants oop. so preallocate cache
    λ_const :: Vector{Float64} #λ=1.0 by default for all callbacks except hess_coord
    
    meta :: NLPModelMeta{Float64, Vector{Float64}}
    counters :: Counters
end

function VarAssimProblem(vavf, vatrc, obs, bds; init=:center)
    callback_module = eval(Meta.parse(callback_module_str(vavf, vatrc)))
    
    (;D, OC, SC, P, NC, R, NNZ_J, NZ_H_PP, NZ_H_M, NZ_H_REM) = callback_module.timestep_params
    (;N) = obs
    
    nvar = (D+OC)*((R+1)*N+1)+P
    
    u_bds, oc_bds, p_bds = bds.u_bds, bds.oc_bds, bds.p_bds

    lb_u, ub_u, init_u = collect.(collect(zip(u_bds...)))
    lb_oc, ub_oc, init_oc = collect.(collect(zip(oc_bds...)))
    lb_p, ub_p, init_p = collect.(collect(zip(p_bds...)))

    lvar = vcat(repeat(vcat(lb_u, lb_oc), (R+1)*N+1), lb_p)
    uvar = vcat(repeat(vcat(ub_u, ub_oc), (R+1)*N+1), ub_p);

    if init == :center
        x0 = (lvar .+ uvar) ./ 2.0
    elseif init == :random
        x0 = lvar .+ (uvar .- lvar) .* (0.47 .+ 0.06 .* rand(length(lvar)))
    elseif init == :user
        x0 = vcat(repeat(vcat(init_u, init_oc), (R+1)*N+1), init_p)
    end
    
    ncon = NC*N
    lcon = zeros(ncon)
    ucon = zeros(ncon);
    
    nnzj = NNZ_J*N
    nnzh = N*(NZ_H_REM+NZ_H_M)+NZ_H_M+NZ_H_PP

    meta = NLPModelMeta(nvar, x0 = x0, 
        lvar = lvar, 
        uvar = uvar, 
        ncon = ncon,
        lcon = lcon,
        ucon = ucon,
        nnzh = nnzh,
        nnzj = nnzj)
    
    counters = Counters()
    
    Obj_cache = [0.0]
    λ_const = ones(ncon)
    
    VarAssimProblem(obs, bds, callback_module,
        Obj_cache, λ_const, 
        meta, counters)
end

function VarAssimProblem(callback_module, obs, bds; init = :center)    
    (;D, OC, SC, P, NC, R, NNZ_J, NZ_H_PP, NZ_H_M, NZ_H_REM) = callback_module.timestep_params
    (;N) = obs
    
    nvar = (D+OC)*((R+1)*N+1)+P
    
    u_bds, oc_bds, p_bds = bds.u_bds, bds.oc_bds, bds.p_bds

    lb_u, ub_u, init_u = collect.(collect(zip(u_bds...)))
    lb_oc, ub_oc, init_oc = collect.(collect(zip(oc_bds...)))
    lb_p, ub_p, init_p = collect.(collect(zip(p_bds...)))

    lvar = vcat(repeat(vcat(lb_u, lb_oc), (R+1)*N+1), lb_p)
    uvar = vcat(repeat(vcat(ub_u, ub_oc), (R+1)*N+1), ub_p);

    if init == :center
        x0 = (lvar .+ uvar) ./ 2.0
    elseif init == :random
        x0 = lvar .+ (uvar .- lvar) .* rand(length(lvar))
    elseif init == :user
        x0 = vcat(repeat(vcat(init_u, init_oc), (R+1)*N+1), init_p)
    end
    
    ncon = NC*N
    lcon = zeros(ncon)
    ucon = zeros(ncon);
    
    nnzj = NNZ_J*N
    nnzh = N*(NZ_H_REM+NZ_H_M)+NZ_H_M+NZ_H_PP

    meta = NLPModelMeta(nvar, x0 = x0, 
        lvar = lvar, 
        uvar = uvar, 
        ncon = ncon,
        lcon = lcon,
        ucon = ucon,
        nnzh = nnzh,
        nnzj = nnzj)
    
    counters = Counters()
    
    Obj_cache = [0.0]
    λ_const = ones(ncon)
    
    VarAssimProblem(obs, bds, callback_module,
        Obj_cache, λ_const, 
        meta, counters)
end

##### Plug in the callbacks to the NLPModels interface

function NLPModels.obj(nlp :: VarAssimProblem, x::AbstractVector)
    nlp.callback_module.cost_fill!(nlp.Obj_cache, x, nlp.λ_const, nlp.obs.SC_rec, nlp.obs.knot_times, nlp.obs.N)
    nlp.Obj_cache[1]
end

function NLPModels.cons!(nlp :: VarAssimProblem, x::AbstractVector, c::AbstractVector)
    nlp.callback_module.cons_fill!(c, x, nlp.λ_const, nlp.obs.SC_rec, nlp.obs.knot_times, nlp.obs.N)
end

function NLPModels.grad!(nlp :: VarAssimProblem, x::AbstractVector, g::AbstractVector)
    nlp.callback_module.grad_fill!(g, x, nlp.λ_const, nlp.obs.SC_rec, nlp.obs.knot_times, nlp.obs.N)
end

function NLPModels.hess_structure!(nlp :: VarAssimProblem, rows::AbstractVector, cols::AbstractVector)
    nlp.callback_module.hess_struct_fill!(rows, cols, nlp.obs.N)
end

function NLPModels.hess_coord!(nlp :: VarAssimProblem, x::AbstractVector, y::AbstractVector, vals::AbstractVector; obj_weight = 1.0)
    nlp.callback_module.hess_fill!(vals, x, y, obj_weight, nlp.obs.SC_rec, nlp.obs.knot_times, nlp.obs.N)
end

function NLPModels.jac_structure!(nlp :: VarAssimProblem, rows::AbstractVector, cols::AbstractVector)
    nlp.callback_module.jac_struct_fill!(rows, cols, nlp.obs.N)
end

function NLPModels.jac_coord!(nlp :: VarAssimProblem, x::AbstractVector, vals::AbstractVector)
    nlp.callback_module.jac_fill!(vals, x, nlp.λ_const, nlp.obs.SC_rec, nlp.obs.knot_times, nlp.obs.N)
end
